 --
   -- Fillable
   -- This is the specialization for fillable vehicles, such as trailers or sprayers
   --
   -- @author  Stefan Geiger
   -- @date  22/07/10
   --
   -- Copyright (C) GIANTS Software GmbH, Confidential, All Rights Reserved.
   
   Fillable = {};
   
   
   Fillable.NUM_FILLTYPES = 0;
   Fillable.FILLTYPE_UNKNOWN = 0;
   Fillable.FILLTYPE_START_TOTAL_AMOUNT = 500000;
   
   Fillable.fillTypeNameToInt = {}
   Fillable.fillTypeIntToName = {}
   
   Fillable.fillTypeIndexToDesc = {}
   Fillable.fillTypeNameToDesc = {}
   Fillable.economyFillTypes = {};
   
   
   Fillable.fillTypeNameToInt["unknown"] = Fillable.FILLTYPE_UNKNOWN;
   Fillable.fillTypeIntToName[Fillable.FILLTYPE_UNKNOWN] = "unknown";
   
   Fillable.sendNumBits = 6;
   
   function Fillable.registerFillType(name, nameI18N, pricePerLiter, partOfEconomy, hudOverlayFilename)
       local key = "FILLTYPE_"..string.upper(name);
       if Fillable[key] == nil then
           if Fillable.NUM_FILLTYPES >= 64 then
               print("Error: Fillable.registerFillType too many fill types. Only 64 fill types are supported");
               return;
           end;
           Fillable.NUM_FILLTYPES = Fillable.NUM_FILLTYPES+1;
           Fillable[key] = Fillable.NUM_FILLTYPES;
           Fillable.fillTypeNameToInt[name] = Fillable.NUM_FILLTYPES;
           Fillable.fillTypeIntToName[Fillable.NUM_FILLTYPES] = name;
   
           local desc = {};
           desc.name = name;
           desc.nameI18N = nameI18N;
           if desc.nameI18N == nil then
               desc.nameI18N = name;
               if g_i18n:hasText(desc.nameI18N) then
                   desc.nameI18N = g_i18n:getText(name);
               end
           end
           desc.index = Fillable.NUM_FILLTYPES;
           desc.pricePerLiter = Utils.getNoNil(pricePerLiter, 0);
           desc.previousHourPrice = desc.pricePerLiter;
           desc.startPricePerLiter = desc.pricePerLiter;
           desc.totalAmount = Fillable.FILLTYPE_START_TOTAL_AMOUNT;
           desc.partOfEconomy = Utils.getNoNil(partOfEconomy, false);
           desc.hudOverlayFilename = hudOverlayFilename;
   
           Fillable.fillTypeIndexToDesc[Fillable.NUM_FILLTYPES] = desc;
           Fillable.fillTypeNameToDesc[name] = desc;
   
           if desc.partOfEconomy then
               table.insert(Fillable.economyFillTypes, desc);
           end;
       end;
       return Fillable[key];
   end;
   
   function Fillable.addFillTypeToEconomy(index)
       if not Fillable.fillTypeIndexToDesc[index].partOfEconomy then
           Fillable.fillTypeIndexToDesc[index].partOfEconomy = true;
           table.insert(Fillable.economyFillTypes, desc);
       end
   end;
   
   function Fillable.prerequisitesPresent(specializations)
       return true;
   end;
   
   function Fillable:load(xmlFile)
   
       self.allowFillType = Fillable.allowFillType;
       self.resetFillLevelIfNeeded = Fillable.resetFillLevelIfNeeded;
       self.setFillLevel = Fillable.setFillLevel;
       self.getFillLevel = Fillable.getFillLevel;
       self.getAllowFillFromAir = Fillable.getAllowFillFromAir;
       self.getFirstEnabledFillType = Fillable.getFirstEnabledFillType;
   
   
       self.fillLevel = 0;
       self.capacity = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.capacity"), 0.0);
       self.fillTypeChangeThreshold = 0.05; -- fill level percentage that still allows overriding with another fill type
   
       self.synchronizeFillLevel = true;
       self.synchronizeFullFillLevel = false;
   
       self.fillRootNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.fillRootNode#index"));
       if self.fillRootNode == nil then
           self.fillRootNode = self.components[1].node;
       end;
   
       self.fillMassNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.fillMassNode#index"));
       if self.fillMassNode == nil then
           self.fillMassNode = self.fillRootNode;
       end;
   
       self.exactFillRootNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.exactFillRootNode#index"));
       if self.exactFillRootNode == nil then
           self.exactFillRootNode = self.fillRootNode;
       end;
   
       self.fillAutoAimTargetNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.fillAutoAimTargetNode#index"));
       if self.fillAutoAimTargetNode == nil then
           self.fillAutoAimTargetNode = self.exactFillRootNode;
       end;
   
       self.fillTypes = {};
       self.fillTypes[Fillable.FILLTYPE_UNKNOWN] = true;
   
       local fillTypes = getXMLString(xmlFile, "vehicle.fillTypes#fillTypes");
       if fillTypes ~= nil then
           local types = Utils.splitString(" ", fillTypes);
           for k,v in pairs(types) do
               local fillType = Fillable.fillTypeNameToInt[v];
               if fillType ~= nil then
                   self.fillTypes[fillType] = true;
               else
                   print("Warning: '"..self.configFileName.. "' has invalid fillType '"..v.."'.");
               end;
           end;
       end;
       local fruitTypes = getXMLString(xmlFile, "vehicle.fillTypes#fruitTypes");
       if fruitTypes ~= nil then
           local types = Utils.splitString(" ", fruitTypes);
           for k,v in pairs(types) do
               local fillType = Fillable.fillTypeNameToInt[v];
               if fillType ~= nil then
                   self.fillTypes[fillType] = true;
               else
                   print("Warning: '"..self.configFileName.. "' has invalid fillType '"..v.."'.");
               end;
           end;
       end;
   
       self.currentFillType = Fillable.FILLTYPE_UNKNOWN;
   
       if self.isServer then
           self.sentFillType = self.currentFillType;
           self.sentFillLevel = self.fillLevel;
       end;
   
       if self.isClient then
           local fillPlanesRotDeg = Utils.getNoNil(getXMLBool(xmlFile, "vehicle.fillPlanes#rotationDegrees"), false);
           self.fillPlanes = {};
           local i = 0;
           while true do
               local key = string.format("vehicle.fillPlanes.fillPlane(%d)", i);
               if not hasXMLProperty(xmlFile, key) then
                   break;
               end;
               local fillPlane = {};
               fillPlane.nodes = {};
               local fillType = getXMLString(xmlFile, key.."#type");
               if fillType ~= nil then
                   local nodeI = 0;
                   while true do
                       local nodeKey = key..string.format(".node(%d)", nodeI);
                       if not hasXMLProperty(xmlFile, nodeKey) then
                           break;
                       end;
                       local node = Utils.indexToObject(self.components, getXMLString(xmlFile, nodeKey.."#index"));
                       if node ~= nil then
                           local defaultX, defaultY, defaultZ = getTranslation(node);
                           local defaultRX, defaultRY, defaultRZ = getRotation(node);
                           setVisibility(node, false);
   
                           local animCurve = AnimCurve:new(linearInterpolatorTransRotScale);
                           local keyI = 0;
                           while true do
                               local animKey = nodeKey..string.format(".key(%d)", keyI);
                               local keyTime = getXMLFloat(xmlFile, animKey.."#time");
                               local x,y,z = Utils.getVectorFromString(getXMLString(xmlFile, animKey.."#translation"));
                               if y == nil then
                                   y = getXMLFloat(xmlFile, animKey.."#y");
                               end;
                               local rx,ry,rz = Utils.getVectorFromString(getXMLString(xmlFile, animKey.."#rotation"));
                               local sx,sy,sz = Utils.getVectorFromString(getXMLString(xmlFile, animKey.."#scale"));
                               if keyTime == nil then
                                   break;
                               end;
                               local x = Utils.getNoNil(x, defaultX);
                               local y = Utils.getNoNil(y, defaultY);
                               local z = Utils.getNoNil(z, defaultZ);
                               if fillPlanesRotDeg then
                                   rx = Utils.getNoNilRad(rx, defaultRX);
                                   ry = Utils.getNoNilRad(ry, defaultRY);
                                   rz = Utils.getNoNilRad(rz, defaultRZ);
                               else
                                   rx = Utils.getNoNil(rx, defaultRX);
                                   ry = Utils.getNoNil(ry, defaultRY);
                                   rz = Utils.getNoNil(rz, defaultRZ);
                               end
                               local sx = Utils.getNoNil(sx, 1);
                               local sy = Utils.getNoNil(sy, 1);
                               local sz = Utils.getNoNil(sz, 1);
                               animCurve:addKeyframe({x=x, y=y, z=z, rx=rx, ry=ry, rz=rz, sx=sx, sy=sy, sz=sz, time = keyTime});
                               keyI = keyI +1;
                           end;
                           if keyI == 0 then
                               local minY, maxY = Utils.getVectorFromString(getXMLString(xmlFile, nodeKey.."#minMaxY"));
                               local minY = Utils.getNoNil(minY, defaultY);
                               local maxY = Utils.getNoNil(maxY, defaultY);
                               animCurve:addKeyframe({x=defaultX, y=minY, z=defaultZ, rx=defaultRX, ry=defaultRY, rz=defaultRZ, sx=1, sy=1, sz=1, time = 0});
                               animCurve:addKeyframe({x=defaultX, y=maxY, z=defaultZ, rx=defaultRX, ry=defaultRY, rz=defaultRZ, sx=1, sy=1, sz=1, time = 1});
                           end;
                           local alwaysVisible = Utils.getNoNil(getXMLBool(xmlFile, nodeKey.."#alwaysVisible"), false);
                           table.insert(fillPlane.nodes, {node=node, animCurve = animCurve, alwaysVisible=alwaysVisible});
                       end;
                       nodeI = nodeI +1;
                   end;
                   if table.getn(fillPlane.nodes) > 0 then
                       if self.defaultFillPlane == nil then
                           self.defaultFillPlane = fillPlane;
                       end;
                       self.fillPlanes[fillType] = fillPlane;
                   end;
               end;
               i = i +1;
           end;
           if self.defaultFillPlane==nil then
               self.fillPlanes = nil;
           end;
   
           if self.fillPlanes == nil then
               Fillable.loadDeprecatedTrailerGrainPlane(self, xmlFile);
               --[[if hasXMLProperty(xmlFile, "vehicle.grainPlane") then
                   print("Warning: '"..self.configFileName.. "' uses old grainPlane format, which is not supported anymore.");
               end;]]
           end;
       end;
   
   
       self.allowFillFromAir = Utils.getNoNil(getXMLBool(xmlFile, "vehicle.allowFillFromAir#value"), true);
   
   
       if self.isServer then
           local shovelFillTriggerId = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.shovelFillTrigger#index"));
           if shovelFillTriggerId ~= nil then
               local shovelFillTrigger = FillableShovelFillTrigger:new();
               if shovelFillTrigger:load(shovelFillTriggerId, self) then
                   self.shovelFillTrigger = shovelFillTrigger;
               else
                   shovelFillTrigger:delete();
               end
           end
       end
   
   
       self.massScale = getXMLFloat(xmlFile, "vehicle.massPerFillLevel#value");
       if self.massScale == nil then
           self.massScale = 1.3*0.0001*0.7 *Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.massScale#value"), 1);
       end
   
       self:setFillLevel(0, Fillable.FILLTYPE_UNKNOWN);
   
   
       setUserAttribute(self.fillRootNode, "vehicleType", "Integer", 2);
   
       self.fillableDirtyFlag = self:getNextDirtyFlag();
   end;
   
   function Fillable:postLoad(xmlFile)
       local startFillLevel = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.startFillLevel"), 0); --self.capacity);
       if startFillLevel > 0 then
           local firstFillType = self:getFirstEnabledFillType();
   
           if firstFillType ~= Fillable.FILLTYPE_UNKNOWN then
               self:setFillLevel(startFillLevel, firstFillType, true);
           end;
       end;
   end;
   
   function Fillable.loadDeprecatedTrailerGrainPlane(self, xmlFile)
   
       --self.grainPlane = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.grainPlane#index"));
       if self.isClient then
           if self.fillPlanes == nil then
               local defaultX, defaultY, defaultZ;
               local defaultRX, defaultRY, defaultRZ;
               local defaultSX, defaultSY, defaultSZ;
   
               self.fillPlanes = {};
               local i = 0;
               while true do
                   local key = string.format("vehicle.grainPlane.node(%d)", i);
                   local fruitType = getXMLString(xmlFile, key.."#type");
                   local index = getXMLString(xmlFile, key.."#index");
                   if fruitType==nil or index==nil then
                       break;
                   end;
                   local node = Utils.indexToObject(self.components, index);
                   if node ~= nil then
                       defaultX, defaultY, defaultZ = getTranslation(node);
                       defaultRX, defaultRY, defaultRZ = getRotation(node);
                       defaultSX, defaultSY, defaultSZ = getScale(node);
   
                       setVisibility(node, false);
                       local fillPlane = {};
                       fillPlane.nodes = {};
                       table.insert(fillPlane.nodes, {node=node, alwaysVisible=false}); -- animCurve is filled below
                       -- Node: fruitType names are equal to the fillType names
                       self.fillPlanes[fruitType] = fillPlane;
   
                       if self.defaultFillPlane == nil then
                           self.defaultFillPlane = fillPlane;
                       end;
                   end;
                   i = i +1;
               end;
               if self.defaultFillPlane==nil then
                   self.fillPlanes = nil;
               else
                   local animCurve = AnimCurve:new(linearInterpolatorTransRotScale);
                   local keyI = 0;
                   while true do
                       local animKey = string.format("vehicle.grainPlane.key(%d)", keyI);
                       local keyTime = getXMLFloat(xmlFile, animKey.."#time");
                       local y = getXMLFloat(xmlFile, animKey.."#y");
                       local sx,sy,sz = Utils.getVectorFromString(getXMLString(xmlFile, animKey.."#scale"));
                       if keyTime == nil then
                           break;
                       end;
                       local x = Utils.getNoNil(x, defaultX);
                       local y = Utils.getNoNil(y, defaultY);
                       local z = Utils.getNoNil(z, defaultZ);
                       local rx = Utils.getNoNil(rx, defaultRX);
                       local ry = Utils.getNoNil(ry, defaultRY);
                       local rz = Utils.getNoNil(rz, defaultRZ);
                       local sx = Utils.getNoNil(sx, defaultSX);
                       local sy = Utils.getNoNil(sy, defaultSY);
                       local sz = Utils.getNoNil(sz, defaultSZ);
                       animCurve:addKeyframe({x=x, y=y, z=z, rx=rx, ry=ry, rz=rz, sx=sx, sy=sy, sz=sz, time = keyTime});
                       keyI = keyI +1;
                   end;
                   if keyI == 0 then
                       local minY, maxY = Utils.getVectorFromString(getXMLString(xmlFile, "vehicle.grainPlane#minMaxY"));
                       local minY = Utils.getNoNil(minY, defaultY);
                       local maxY = Utils.getNoNil(maxY, defaultY);
                       animCurve:addKeyframe({x=defaultX, y=minY, z=defaultZ, rx=defaultRX, ry=defaultRY, rz=defaultRZ, sx=defaultSX, sy=defaultSY, sz=defaultSZ, time = 0});
                       animCurve:addKeyframe({x=defaultX, y=maxY, z=defaultZ, rx=defaultRX, ry=defaultRY, rz=defaultRZ, sx=defaultSX, sy=defaultSY, sz=defaultSZ, time = 1});
                   end;
   
                   for _, fillPlane in pairs(self.fillPlanes) do
                       fillPlane.nodes[1].animCurve = animCurve;
                   end;
               end;
           end;
       end;
   
   end;
   
   function Fillable:delete()
       if self.shovelFillTrigger ~= nil then
           self.shovelFillTrigger:delete();
       end
   end;
   
   function Fillable:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
       if self.synchronizeFillLevel then
           local fillLevel = getXMLFloat(xmlFile, key.."#fillLevel");
           local fillType = getXMLString(xmlFile, key.."#fillType");
           if fillLevel ~= nil and fillType ~= nil then
               local fillTypeInt = Fillable.fillTypeNameToInt[fillType];
               if fillTypeInt ~= nil then
                   self:setFillLevel(fillLevel, fillTypeInt);
               end;
           end;
       end
       return BaseMission.VEHICLE_LOAD_OK;
   end;
   
   function Fillable:getSaveAttributesAndNodes(nodeIdent)
       if self.synchronizeFillLevel then
           local fillType = Fillable.fillTypeIntToName[self.currentFillType];
           if fillType == nil then
               fillType = "unknown";
           end;
           local attributes = 'fillLevel="'..self.fillLevel..'" fillType="'..fillType..'"';
           return attributes, nil;
       end
       return nil;
   end;
   
   function Fillable:addNodeVehicleMapping(list)
       list[self.fillRootNode] = self;
       list[self.exactFillRootNode] = self;
   end;
   
   function Fillable:removeNodeVehicleMapping(list)
       list[self.fillRootNode] = nil;
       list[self.exactFillRootNode] = nil;
   end;
   
   function Fillable:mouseEvent(posX, posY, isDown, isUp, button)
   end;
   
   function Fillable:keyEvent(unicode, sym, modifier, isDown)
   end;
   
   function Fillable:readStream(streamId, connection)
       if connection:getIsServer() then
           if self.synchronizeFillLevel then
               local fillLevel = streamReadFloat32(streamId);
               local fillType = streamReadUIntN(streamId, Fillable.sendNumBits);
               self:setFillLevel(fillLevel, fillType);
           end
       end
   end;
   
   function Fillable:writeStream(streamId, connection)
       if not connection:getIsServer() then
           if self.synchronizeFillLevel then
               streamWriteFloat32(streamId, self.fillLevel);
               streamWriteUIntN(streamId, self.currentFillType, Fillable.sendNumBits);
           end
       end;
   end;
   
   function Fillable:readUpdateStream(streamId, timestamp, connection)
       if connection:getIsServer() then
           if self.synchronizeFillLevel then
               if streamReadBool(streamId) then
                   local fillLevel;
                   if self.synchronizeFullFillLevel then
                       fillLevel = streamReadFloat32(streamId);
                   else
                       fillLevel = streamReadUInt16(streamId)/65535*self.capacity;
                   end
                   local fillType = streamReadUIntN(streamId, Fillable.sendNumBits);
   
                   self:setFillLevel(fillLevel, fillType, true);
               end;
           end
       end;
   end;
   
   function Fillable:writeUpdateStream(streamId, connection, dirtyMask)
       if not connection:getIsServer() then
           if self.synchronizeFillLevel then
               if streamWriteBool(streamId, bitAND(dirtyMask, self.fillableDirtyFlag) ~= 0) then
                   if self.synchronizeFullFillLevel then
                       streamWriteFloat32(streamId, self.fillLevel);
                   else
                       local percent = 0;
                       if self.capacity ~= 0 then
                           percent = Utils.clamp(self.fillLevel / self.capacity, 0, 1);
                       end;
                       streamWriteUInt16(streamId, math.floor(percent*65535));
                   end
                   streamWriteUIntN(streamId, self.currentFillType, Fillable.sendNumBits);
               end;
           end
       end;
   end;
   
   function Fillable:update(dt)
       if self.firstTimeRun then
           if self.isServer then
               if self.emptyMass == nil then
                   self.emptyMass = getMass(self.fillMassNode);
                   self.currentMass = self.emptyMass;
               end;
               local newMass = self.emptyMass + self.fillLevel*self.massScale;
               if newMass ~= self.currentMass then
                   setMass(self.fillMassNode, newMass);
                   self.currentMass = newMass;
                   --[[for k,v in pairs(self.components) do
                       if v.node == self.fillMassNode then
                           if v.centerOfMass ~= nil then
                               setCenterOfMass(v.node, v.centerOfMass[1], v.centerOfMass[2], v.centerOfMass[3]);
                           end;
                           break;
                       end;
                   end;]]
               end;
           end;
       end;
   end;
   
   function Fillable:updateTick(dt)
       if self.isServer then
           if self.synchronizeFillLevel then
               if self.fillLevel ~= self.sentFillLevel or self.currentFillType ~= self.sentFillType then
                   self:raiseDirtyFlags(self.fillableDirtyFlag);
   
                   self.sentFillLevel = self.fillLevel;
                   self.sentFillType = self.currentFillType;
               end;
           end
           if self.shovelFillTrigger ~= nil then
               self.shovelFillTrigger:update(dt);
           end
       end;
   end;
   
   function Fillable:draw()
   end;
   
   function Fillable:resetFillLevelIfNeeded(fillType)
       if self.currentFillType ~= fillType then
           self.fillLevel = 0;
       end;
   end;
   
   function Fillable:allowFillType(fillType, allowEmptying)
       local allowed = false;
   
       if self.fillTypes[fillType] then
           if self.currentFillType ~= Fillable.FILLTYPE_UNKNOWN then
               if self.currentFillType ~= fillType then
                   if self.fillLevel <= self.capacity*self.fillTypeChangeThreshold then
                       allowed = true; -- fill level is low enough to be overridden
                       if allowEmptying then
                           self.fillLevel = 0; -- empty the trailer
                       end;
                   end;
               else
                   allowed = true; -- fill type is the same as the trailer's current fill type
               end;
           else
               allowed = true; -- fillable is empty --> Fillable.FILLTYPE_UNKNOWN
           end;
       end;
   
       return allowed;
   end;
   
   function Fillable:getFillLevel(fillType)
       if fillType == self.currentFillType then
           return self.fillLevel;
       end
       return 0;
   end
   
   function Fillable:setFillLevel(fillLevel, fillType, force)
       if (force == nil or force == false) and not self:allowFillType(fillType, false) then
           return
       end;
   
       self.currentFillType = fillType;
       self.fillLevel = fillLevel;
       if self.fillLevel > self.capacity then
           self.fillLevel = self.capacity;
       end;
       if self.fillLevel <= 0 then
           self.fillLevel = 0;
           self.currentFillType = Fillable.FILLTYPE_UNKNOWN;
       end;
   
   
       if self.isClient then
           if self.currentFillPlane ~= nil then
               for _, node in ipairs(self.currentFillPlane.nodes) do
                   setVisibility(node.node, false);
               end;
               self.currentFillPlane = nil;
           end;
           if self.fillPlanes ~= nil and self.defaultFillPlane ~= nil then
               local fillPlane;
               local t = 0;
               if fillType ~= Fillable.FILLTYPE_UNKNOWN then
                   local fillTypeName = Fillable.fillTypeIntToName[fillType];
                   fillPlane = self.fillPlanes[fillTypeName];
                   t = self.fillLevel/self.capacity;
               end
               if fillPlane == nil then
                   fillPlane = self.defaultFillPlane
               end
               for _, node in ipairs(fillPlane.nodes) do
                   local x,y,z, rx,ry,rz, sx,sy,sz = node.animCurve:get(t);
   
                   setTranslation(node.node, x, y, z);
                   setRotation(node.node, rx, ry, rz);
                   setScale(node.node, sx, sy, sz);
                   setVisibility(node.node, self.fillLevel > 0 or node.alwaysVisible);
               end;
               self.currentFillPlane = fillPlane;
           end;
       end;
   end;
   
   function Fillable:getAllowFillFromAir()
       return self.allowFillFromAir;
   end;
   
   function Fillable:getFirstEnabledFillType()
       for fillType, enabled in pairs(self.fillTypes) do
           if fillType ~= Fillable.FILLTYPE_UNKNOWN and enabled then
               return fillType;
           end
       end
       return Fillable.FILLTYPE_UNKNOWN;
   end